/*
 * Gnutmeg - a peer to peer distributed file sharing system.
 * 
 * Copyright (c) 2000 Dave Dunkin
 * Released under the GNU Public License, see COPYING
 *
 * $id:
 */

import java.io.*;
import java.net.*;
import java.util.*;

import gnu.regexp.*;

public class Gnutmeg
{
	public final static boolean TRACE          = false;

	public final static byte    GNUTMEG_VER    = 1;  //protocol version
	public final static int     DEFAULT_PORT   = 9911;

	public final static int     MAX_MSG_LENGTH = 1024;  //max length for datagram payload

	//packet identifiers
	public final static byte    PKT_HELLO      = 1;  //request for contacts (alerts receiver that sender is online)
	public final static byte    PKT_CONTACT    = 2;  //response to HELLO, contains contacts
	public final static byte    PKT_SEARCH     = 3;  //request to search, contains search pattern
	public final static byte    PKT_FOUND      = 4;  //response to SEARCH, contains results
	public final static byte    PKT_SEND       = 5;  //request to initiate a transfer, contains name of requested file
	public final static byte    PKT_PING       = 6;  //request to check if host is alive
	public final static byte    PKT_PONG       = 7;  //response indicating host is alive, contains linespeed info

	public final static int     LS_UNKNOWN     = 0;
	public final static int     LS_LOW         = 1;  //14.4-33.6Kbps
	public final static int     LS_MEDIUM      = 2;  //56K, ISDN
	public final static int     LS_FAST        = 3;  //T1, cable, dsl

	private Vector         m_contacts      = null;
	private Vector         m_searches      = null;
	private Vector         m_searchResults = null;
	private Vector         m_downloads     = null;
	private String[]       m_filelist      = null;

	private int            m_linespeed     = LS_UNKNOWN;
	private int            m_maxContacts   = (m_linespeed+1)*10;
	private int            m_searchHist    = 100;
	private int            m_port          = DEFAULT_PORT;
	private String         m_filebase      = ".";
	private int            m_nextSearchId  = (int)System.currentTimeMillis();
	private byte           m_searchTTL     = 16;

	private boolean        m_listening     = true;

	private DatagramSocket m_socket        = null;

	public Gnutmeg()
	{
		m_contacts = new Vector(m_maxContacts);
		m_searches = new Vector(m_searchHist);
		m_searchResults = new Vector();
		m_downloads = new Vector();

		//build filelist
		m_filelist = listFiles(m_filebase);

		//bind socket
		try
		{
			m_socket = new DatagramSocket(m_port, InetAddress.getLocalHost());
			System.out.println("Bound to "+m_socket.getLocalAddress()+':'+m_socket.getLocalPort());
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}


		//start listener thread
		Thread socketListener = new Thread()
		{
			public void run()
			{
				while (m_listening)
				{
					try
					{
						byte[] buff = new byte[MAX_MSG_LENGTH];
						DatagramPacket packet = new DatagramPacket(buff, buff.length);
						m_socket.receive(packet);

						//TODO: put the packet in a queue and alert another thread to handle it

						byte[] data = packet.getData();
						if (TRACE) System.err.println("Got packet with id: "+data[1]);
						if (data[0] == GNUTMEG_VER)
						{
							switch (data[1])
							{
								case PKT_HELLO:
									processPktHello(packet);
									break;
								case PKT_CONTACT:
									processPktContact(packet);
									break;
								case PKT_SEARCH:
									processPktSearch(packet);
									break;
								case PKT_FOUND:
									processPktFound(packet);
									break;
								case PKT_SEND:
									processPktSend(packet);
									break;
								default:
									//error in communications
									if (TRACE) System.err.println("MALFORMED PACKET RECEIVED");
							}
						}
						else
						{
							if (TRACE) System.err.println("Wrong protocol version: "+data[0]);
						}
					}
					catch (SocketException ex)
					{
					}
					catch (IOException ex)
					{
						ex.printStackTrace();
					}
				}
			}
		};
		socketListener.start();
	}

	/**
	 * Add contact to list. Should be of the form hostname:port.
	 */
	public void addContact(String contact)
	{
		try
		{
			addContact(new Contact(contact));
		}
		catch (UnknownHostException ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Add contact to list. Should be of the form hostname:port.
	 */
	public void addContact(String contact, int linespeed)
	{
		try
		{
			addContact(new Contact(contact, linespeed));
		}
		catch (UnknownHostException ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Add contact to list.
	 */
	public void addContact(Contact contact)
	{
		if (TRACE) System.err.println("addContact("+contact+")");

		try
		{
			if (!isLocal(contact.addr, contact.port))
			{
				if (!m_contacts.contains(contact))
				{
					byte[] data = {GNUTMEG_VER, PKT_HELLO, (byte)m_linespeed};
					contact.send(data);
				}
			}
		}
		catch (SocketException ex)
		{
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
	}

	public void removeContact(String contact)
	{
		try
		{
			m_contacts.removeElement(new Contact(contact));
		}
		catch (UnknownHostException ex)
		{
			//FIXME: should probably check another way
		}
	}

	public void search(String pattern)
	{
		m_searchResults.removeAllElements();

		Search search = new Search();
		search.pattern = pattern;
		search.id = m_nextSearchId++;
		search.ttl = m_searchTTL;
		search.contact = new Contact(m_socket.getLocalAddress(), m_port);

		sendSearch(search);
	}

	public void get(final String file, Contact contact)
	{
		//find a tcp port to listen on
		final int port = m_port+1;
		Thread listenerThread = new Thread()
		{
			public void run()
			{
				if (TRACE) System.err.println("Waiting for: "+file);
				if (TRACE) System.err.println(" on port: "+port);

				String fname = file.substring(file.lastIndexOf('/')+1);
				try
				{
					ServerSocket serverSocket = new ServerSocket(port);
					Socket socket = serverSocket.accept();
					serverSocket.close();  //no new connections
					InputStream in = new BufferedInputStream(socket.getInputStream());
					BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(fname));

					if (TRACE) System.err.println("Transfer beginning: "+file);

					int b = in.read();
					while (b != -1)
					{
						out.write(b);
						b = in.read();
					}
					out.flush();
					socket.close();

					if (TRACE) System.err.println("Transfer completed: "+file);
				}
				catch (IOException ex)
				{
					ex.printStackTrace();
				}
			}
		};
		m_downloads.addElement(listenerThread);
		listenerThread.start();

		try
		{
			ByteArrayOutputStream bstream = new ByteArrayOutputStream(MAX_MSG_LENGTH);
			DataOutputStream stream = new DataOutputStream(bstream);
			stream.writeByte(GNUTMEG_VER);
			stream.writeByte(PKT_SEND);

			stream.writeShort(port);
			stream.writeBytes(file);

			stream.flush();
			contact.send(bstream.toByteArray());
		}
		catch (SocketException ex)
		{
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}

		m_downloads.removeElement(listenerThread);
	}

	public Vector getSearchResults()
	{
		return m_searchResults;
	}

	public Vector getContacts()
	{
		return m_contacts;
	}

	public String[] getFileList()
	{
		return m_filelist;
	}

	public void destroy()
	{
		m_listening = false;
		m_socket.close();
	}

	public static void main(String[] args)
	throws Exception
	{
		Gnutmeg nutmeg = new Gnutmeg();

		LineNumberReader in = new LineNumberReader(new InputStreamReader(System.in));
		boolean quit = false;
		while (!quit)
		{
			System.out.println();
			System.out.print("> ");
			String line = in.readLine();

			StringTokenizer tokens = new StringTokenizer(line);
			String command = tokens.hasMoreTokens() ? tokens.nextToken() : "";
			if (command.equals("search"))
			{
				nutmeg.search(line.substring(7));
			}
			else if (command.equals("results"))
			{
				System.out.println("Search results:");
				int i = 1;
				Enumeration e = nutmeg.getSearchResults().elements();
				while (e.hasMoreElements())
				{
					System.out.println((i++)+". "+e.nextElement());
				}
			}
			else if (command.equals("files"))
			{
				System.out.println("Shared files:");	
				String[] filelist = nutmeg.getFileList();
				for (int i = 0; i < filelist.length; i++)
				{
					System.out.println(filelist[i]);
				}
			}
			else if (command.equals("contacts"))
			{
				System.out.println("Contacts:");	
				Enumeration e = nutmeg.getContacts().elements();
				while (e.hasMoreElements())
				{
					System.out.println(e.nextElement());
				}
			}
			else if (command.equals("add"))
			{
				while (tokens.hasMoreTokens())
				{
					nutmeg.addContact(tokens.nextToken());
				}
			}
			else if (command.equals("remove"))
			{
				while (tokens.hasMoreTokens())
				{
					nutmeg.removeContact(tokens.nextToken());
				}
			}
			else if (command.equals("get"))
			{
				Vector searchResults = nutmeg.getSearchResults();
				while (tokens.hasMoreTokens())
				{
					int i = Integer.valueOf(tokens.nextToken()).intValue();
					SearchResult sr = (SearchResult) searchResults.elementAt(i-1);
					nutmeg.get(sr.file, sr.contact);
				}
			}
			else if (command.equals("help"))
			{
				System.out.println("search");
				System.out.println("results");
				System.out.println("contacts");
				System.out.println("add");
				System.out.println("remove");
				System.out.println("get");
				System.out.println("help");
				System.out.println("quit");
			}
			else if (command.equals("quit"))
			{
				quit = true;
				nutmeg.destroy();
			}

		}
		
	}


////////////////////////////////////////////////////////////
// private functions

	private void sendSearch(Search search)
	{
		if (TRACE) System.err.println("Sending search: "+search.pattern);

		m_searches.addElement(search);

		Enumeration e = m_contacts.elements();
		while (e.hasMoreElements())
		{
			try
			{
				Contact contact = (Contact) e.nextElement();

				if (!contact.equals(search.contact))
				{
					ByteArrayOutputStream bstream = new ByteArrayOutputStream(MAX_MSG_LENGTH);
					DataOutputStream stream = new DataOutputStream(bstream);
					stream.writeByte(GNUTMEG_VER);
					stream.writeByte(PKT_SEARCH);

					stream.writeInt(search.id);
					stream.writeByte(search.ttl);

					stream.write(search.contact.addr.getAddress(), 0, 4);  //not IPv6 compatible
					stream.writeShort((short)search.contact.port);

					stream.writeBytes(search.pattern);

					stream.flush();
					contact.send(bstream.toByteArray());
				}
			}
			catch (SocketException ex)
			{
			}
			catch (IOException ex)
			{
				ex.printStackTrace();
			}
		}
	}

	private void sendContacts(Contact contact)
	{
		if (TRACE) System.err.println("sendContacts("+contact+')');

		try
		{
			ByteArrayOutputStream bstream = new ByteArrayOutputStream(MAX_MSG_LENGTH);
			DataOutputStream stream = new DataOutputStream(bstream);
			stream.writeByte(GNUTMEG_VER);
			stream.writeByte(PKT_CONTACT);

			//TODO: instead of sending all contacts, send a random sampling
			int prev = 0;
			Enumeration e = m_contacts.elements();
			while (e.hasMoreElements())
			{
				Contact c = (Contact) e.nextElement();

				if ((stream.size()-prev) + 8 > MAX_MSG_LENGTH)
				{
					//send this packet now, it's full
					stream.flush();
					prev += stream.size();  //size() doesn't reset with flush()
					contact.send(bstream.toByteArray());

					bstream.reset();
					stream.writeByte(GNUTMEG_VER);
					stream.writeByte(PKT_CONTACT);
				}

				stream.writeByte(GNUTMEG_VER);
				stream.write(c.addr.getAddress(), 0, 4);  //not IPv6 compatible
				stream.writeShort((short)c.port);
				stream.writeByte((byte)c.linespeed);
			}

			stream.flush();
			contact.send(bstream.toByteArray());
		}
		catch (SocketException ex)
		{
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
	}

	private boolean isLocal(InetAddress addr, int port)
	{
		return addr.equals(m_socket.getLocalAddress()) && port == m_socket.getLocalPort();
	}


////////////////////////////////////////////////////////////
// process packets

	private void processPktHello(DatagramPacket packet)
	{
		byte[] data = packet.getData();
		if (TRACE) System.err.println("processPktHello: "/*+(new String(data))*/);

		try
		{
			DataInputStream stream = new DataInputStream(new ByteArrayInputStream(data));
			stream.skipBytes(2);  //GNUTMEG_VER, PKT_HELLO

			byte linespeed = stream.readByte();
			Contact contact = new Contact(packet.getAddress(), packet.getPort(), linespeed);
			sendContacts(contact);
			if (!m_contacts.contains(contact) &&
				m_contacts.size() <= m_maxContacts &&
				!isLocal(contact.addr, contact.port))
			{
				m_contacts.addElement(contact);
			}
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}

	}

	private void processPktContact(DatagramPacket packet)
	{
		byte[] data = packet.getData();
		if (TRACE) System.err.println("processPktContact: "/*+(new String(data))*/);

		try
		{
			DataInputStream stream = new DataInputStream(new ByteArrayInputStream(data));
			stream.skipBytes(2);  //GNUTMEG_VER, PKT_CONTACT

			int linespeed = stream.readByte();
			Contact contact = new Contact(packet.getAddress(), packet.getPort(), linespeed);
			if (!m_contacts.contains(contact) &&
				m_contacts.size() <= m_maxContacts &&
				!isLocal(contact.addr, contact.port))
			{
				m_contacts.addElement(contact);
			}

			while (stream.available() > 8)
			{
				byte ver = stream.readByte();
				if (ver == GNUTMEG_VER)
				{
					byte[] addrBytes = new byte[4];
					stream.read(addrBytes);
					int port = ushortToInt(stream.readShort());
					linespeed = stream.readByte();
					addContact(new Contact(InetAddress.getByName(ipToString(addrBytes)), port, linespeed));
				}
				else
				{
					stream.skipBytes(7);
				}
			}
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
	}

	private void processPktSearch(DatagramPacket packet)
	{
		byte[] data = packet.getData();
		if (TRACE) System.err.println("processPktSearch: "/*+(new String(data))*/);

		try
		{
			DataInputStream stream = new DataInputStream(new ByteArrayInputStream(data));
			stream.skipBytes(2);  //GNUTMEG_VER, PKT_SEARCH

			int id = stream.readInt();
			byte ttl = stream.readByte();
			byte[] addrBytes = new byte[4];
			stream.read(addrBytes);
			int port = ushortToInt(stream.readShort());
			byte[] patternBytes = new byte[data.length-13];
			stream.readFully(patternBytes);
			String pattern = new String(patternBytes).trim();

			Search search = new Search();
			search.id = id;
			search.ttl = --ttl;
			search.pattern = pattern;
			search.contact = new Contact(InetAddress.getByName(ipToString(addrBytes)), port);

			if (TRACE) System.err.println("Got search from: "+search.contact);

			if (!m_searches.contains(search))
			{
				if (TRACE) System.err.println("Searching for: "+search.pattern);

				//do the searching
				RE regex = new RE(search.pattern);
				Vector results = new Vector();
				for (int i = 0; i < m_filelist.length; i++)
				{
					if (regex.getMatch(m_filelist[i]) != null)
						results.addElement(m_filelist[i]);
				}

				//pass it on
				if (search.ttl > 0)
					sendSearch(search);

				if (results.size() > 0)
				{
					ByteArrayOutputStream bstream = new ByteArrayOutputStream(MAX_MSG_LENGTH);
					DataOutputStream ostream = new DataOutputStream(bstream);
					ostream.writeByte(GNUTMEG_VER);
					ostream.writeByte(PKT_FOUND);
					ostream.writeByte(m_linespeed);

					int prev = 0;
					Enumeration e = results.elements();
					while (e.hasMoreElements())
					{
						String result = (String) e.nextElement();

						if ((ostream.size()-prev) + result.length() > MAX_MSG_LENGTH)
						{
							//send this packet now, it's full
							ostream.flush();
							prev += ostream.size();  //size() doesn't reset with flush()
							search.contact.send(bstream.toByteArray());

							bstream.reset();
							ostream.writeByte(GNUTMEG_VER);
							ostream.writeByte(PKT_FOUND);
							ostream.writeByte(m_linespeed);
						}

						ostream.writeBytes(result+' ');
					}

					ostream.flush();
					search.contact.send(bstream.toByteArray());
				}
			}
		}
		catch (SocketException ex)
		{
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
		catch (REException ex)
		{
			ex.printStackTrace();
		}
	}

	private void processPktFound(DatagramPacket packet)
	{
		byte[] data = packet.getData();
		if (TRACE) System.err.println("processPktFound: "+(new String(data)));

		int linespeed = data[2];

		boolean quoted = false;
		StringBuffer qbuffer = null;
		StringTokenizer tokens = new StringTokenizer(new String(data, 3, data.length-3));
		while (tokens.hasMoreTokens())
		{
			String result = tokens.nextToken().trim();

			if (result.indexOf('\"') == 0)
			{
				quoted = true;
				qbuffer = new StringBuffer(result);
			}
			else if (quoted)
			{
				qbuffer.append(' ');
				qbuffer.append(result);
				if (result.indexOf('\"') == result.length()-1 &&
				    result.charAt(result.length()-2) != '\\')
				{
					quoted = false;
					SearchResult sr = new SearchResult();
					sr.file = qbuffer.toString();
					sr.contact = new Contact(packet.getAddress(), packet.getPort(), linespeed);
					m_searchResults.addElement(sr);
				}
			}
			else if (result.length() > 0)
			{
				SearchResult sr = new SearchResult();
				sr.file = result;
				sr.contact = new Contact(packet.getAddress(), packet.getPort(), linespeed);
				m_searchResults.addElement(sr);
			}
		}
	}

	private void processPktSend(DatagramPacket packet)
	{
		byte[] data = packet.getData();
		if (TRACE) System.err.println("processPktSend: "/*+(new String(data))*/);

		try
		{
			DataInputStream stream = new DataInputStream(new ByteArrayInputStream(data));
			stream.skipBytes(2);  //GNUTMEG_VER, PKT_SEND

			int port = ushortToInt(stream.readShort());
			byte[] fileBytes = new byte[data.length-4];
			stream.readFully(fileBytes);
			String file = unQuoteString(new String(fileBytes).trim());

			if (TRACE) System.err.println("Opening socket to: "+packet.getAddress()+':'+port);

			Socket socket = new Socket(packet.getAddress(), port);
			BufferedOutputStream out = new BufferedOutputStream(socket.getOutputStream());
			InputStream in = new BufferedInputStream(new FileInputStream(m_filebase+file));

			if (TRACE) System.err.println("Transfer beginning: "+file);

			int b = in.read();
			while (b != -1)
			{
				out.write(b);
				b = in.read();
			}
			out.flush();
			socket.close();

			if (TRACE) System.err.println("Transfer completed: "+file);
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
	}


////////////////////////////////////////////////////////////
// static helper methods

	private static String ipToString(byte[] addrBytes)
	{
		return new StringBuffer().append(ubyteToInt(addrBytes[0])).append('.').append(ubyteToInt(addrBytes[1])).append('.').append(ubyteToInt(addrBytes[2])).append('.').append(ubyteToInt(addrBytes[3])).toString();
	}

	private static int ubyteToInt(byte b)
	{
		return (b+256)%256;
	}

	private static int ushortToInt(short s)
	{
		return (s+65536)%65536;
	}

	private static void listFilesRec(File dir, String path, Vector files)
	{
		String[] list = dir.list();
		for (int i = 0; i < list.length; i++)
		{
			File f = new File(path, list[i]);
			if (f.isDirectory())
				listFilesRec(f, path+'/'+list[i], files);
			else
				files.addElement(path+'/'+list[i]);
		}
	}

	private static String[] listFiles(String path)
	{
		File basedir = new File(path);
		Vector files = new Vector();
		listFilesRec(basedir, path, files);

		String[] list = new String[files.size()];
		int i = 0;
		Enumeration e = files.elements();
		while (e.hasMoreElements())
		{
			String fname = ((String) e.nextElement()).substring(path.length());
			fname = quoteString(fname);
			list[i++] = fname;
		}

		return list;
	}

	private static String quoteString(String str)
	{
		int i = str.indexOf('\"');
		while (i > -1)
		{
			str = str.substring(0, i)+'\\'+str.substring(i);
			i = str.indexOf('\"', i+2);  //we inserted a char
		}

		i = str.indexOf('\\');
		while (i > -1)
		{
			str = str.substring(0, i)+'\\'+str.substring(i);
			i = str.indexOf('\\', i+2);  //we inserted a char
		}

		if (str.indexOf(' ') > -1)
			str = '\"'+str+'\"';

		return str;
	}

	private static String unQuoteString(String str)
	{
		if (str.charAt(0) == '\"') // && str.charAt(str.length()-1) == '\"')
			str = str.substring(1, str.length()-1);

		int i = str.indexOf('\"');
		while (i > -1)
		{
			str = str.substring(0, i-1)+str.substring(i);
			i = str.indexOf('\"', i);  //we removed a char
		}

		return str;
	}


////////////////////////////////////////////////////////////
// inner classes

	private class Search
	{
		Contact contact;
		int     id;
		byte    ttl;
		String  pattern;

		public boolean equals(Object obj)
		{
			return obj instanceof Search &&
			       ((Search)obj).contact.equals(contact) &&
			       ((Search)obj).id == id;
		}
	}

	private class SearchResult
	{
		Contact contact;
		String  file;

		public String toString()
		{
			return contact.toString()+' '+file;
		}
	}

	private class Contact
	{
		InetAddress addr;
		int         port;
		int         linespeed;

		public Contact(InetAddress addr, int port)
		{
			this.addr = addr;
			this.port = port;
			this.linespeed = LS_UNKNOWN;
		}

		public Contact(InetAddress addr, int port, int linespeed)
		{
			this.addr = addr;
			this.port = port;
			this.linespeed = linespeed;
		}

		/**
		 * @param contact Of the form hostname:port
		 */
		public Contact(String contact)
		throws UnknownHostException
		{
			int i = contact.indexOf(':');
			if (i == -1)
			{
				addr = InetAddress.getByName(contact);
				port = DEFAULT_PORT;
				linespeed = LS_UNKNOWN;
			}
			else
			{
				String host = contact.substring(0, i);
				addr = InetAddress.getByName(host);
				port = Integer.valueOf(contact.substring(i+1)).intValue();
				linespeed = LS_UNKNOWN;
			}
		}

		/**
		 * @param contact Of the form hostname:port
		 */
		public Contact(String contact, int linespeed)
		throws UnknownHostException
		{
			int i = contact.indexOf(':');
			if (i == -1)
			{
				addr = InetAddress.getByName(contact);
				port = DEFAULT_PORT;
				linespeed = linespeed;
			}
			else
			{
				String host = contact.substring(0, i);
				addr = InetAddress.getByName(host);
				port = Integer.valueOf(contact.substring(i+1)).intValue();
				linespeed = linespeed;
			}
		}

		public void send(byte[] data)
		throws IOException
		{
			if (TRACE) System.err.println("Sending packet of id "+data[1]+" to: "+toString());
			DatagramPacket packet = new DatagramPacket(data, data.length, addr, port);
			m_socket.send(packet);
		}

		public boolean equals(Object obj)
		{
			return obj instanceof Contact &&
			       ((Contact)obj).addr.equals(addr) &&
			       ((Contact)obj).port == port;
		}

		public String toString()
		{
			return addr.getHostAddress()+':'+port;
		}
	}
}
